import Base: findfirst

struct Near{TTarget,TTolerance}
    target::TTarget
    tolerance::TTolerance
end

Near(target) = Near(target,nothing)


function _findfirst(n::Near,v::AbstractVector)
    diff, ind = findmin(x -> abs(x - n.target),v)

    if (n.tolerance != nothing) && (diff > n.tolerance)
        return Int[]
    else
        return ind
    end
end

function scan_exp!(exp::Symbol,found)
    if haskey(found,exp)
        return found[exp]
    else
        newsym = gensym()
        found[exp] = newsym
        return newsym
    end
end

function scan_exp!(exp::Expr,found)
    if exp.head == :$
        return exp.args[1]
    end

    if exp.head == :call
        # skip function name
        return Expr(exp.head,exp.args[1],scan_exp!.(exp.args[2:end],Ref(found))...)
    elseif exp.head == :comparison
        # :(3 <= lon <= 7.2)
        args = Vector{Any}(undef,length(exp.args))
        for i = 1:length(exp.args)
            if iseven(i)
                # skip inflix operators
                args[i] = exp.args[i]
            else
                args[i] = scan_exp!(exp.args[i],found)
            end
        end
        return Expr(exp.head,args...)
    else
        return Expr(exp.head,scan_exp!.(exp.args[1:end],Ref(found))...)
    end
end

# neither Expr nor Symbol
scan_exp!(exp,found) = exp

function scan_exp(exp::Expr)
    found = Dict{Symbol,Symbol}()
    exp = scan_exp!(exp,found)
    return found,exp
end

function scan_coordinate_name(exp)
    params,exp = scan_exp(exp)

    if length(params) != 1
        error("Multiple (or none) coordinates in expression $exp ($params) while looking for $(coordinate_names).")
    end
    param = first(params)
    return param,exp
end


function split_by_and!(exp,sub_exp)
    if exp.head == :&&
        split_by_and!(exp.args[1],sub_exp)
        split_by_and!(exp.args[2],sub_exp)
    else
        push!(sub_exp,exp)
    end
    return sub_exp
end

split_by_and(exp) = split_by_and!(exp,[])

_intersect(r1::AbstractVector,r2::AbstractVector) = intersect(r1,r2)
_intersect(r1::AbstractVector,r2::Number) = (r2 in r1 ? r2 : [])
_intersect(r1::Number,r2::Number) = (r2 == r1 ? r2 : [])
_intersect(r1::Colon,r2) = r2
_intersect(r1::Colon,r2::AbstractRange) = r2





"""
    vsubset = CommonDataModel.@select(v,expression)
    dssubset = CommonDataModel.@select(ds,expression)

Return a subset of the variable `v` (or dataset `ds`) satisfying the condition `expression` as a view. The condition
has the following form:

`condition₁ && condition₂ && condition₃ ... conditionₙ`

Every condition should involve a single 1D variable (typically a coordinate variable, referred as `coord` below). If `v`
is a variable, the related 1D variable should have a shared dimension with the variable `v`. All local variables need to have a `\$` prefix (see examples below). This macro is experimental and subjected to change.

Every condition can either perform:

* a nearest match: `coord ≈ target_coord` (for `≈` type `\\approx` followed by the TAB-key). Only the data corresponding to the index closest to `target_coord` is loaded.

* a nearest match with tolerance: `coord ≈ target_coord ± tolerance`. As before, but if the difference between the closest value in `coord` and `target_coord` is larger (in absolute value) than `tolerance`, an empty array is returned.

* a condition operating on scalar values. For example, a `condition` equal to `10 <= lon <= 20` loads all data with the longitude between 10 and 20 or `abs(lat) > 60` loads all variables with a latitude north of 60° N and south of 60° S (assuming that the has the 1D variables `lon` and `lat` for longitude and latitude).

Only the data which satisfies all conditions is loaded. All conditions must be chained with an `&&` (logical and). They should not contain additional parenthesis or other logical operators such as `||` (logical or).

To convert the view into a regular array one can use `collect`, `Array` or regular indexing.
As in julia, views of scalars are wrapped into a zero dimensional arrays which can be dereferenced by using `[]`. Modifying a view will modify the underlying file (if
the file is opened as writable, otherwise an error is issued).

As for any view, one can use `parentindices(vsubset)` to get the indices matching a select query.

## Examples

Create a sample file with random data:

```julia
using NCDatasets, Dates
using CommonDataModel: @select
# or
# using NCDatasets: @select

fname = "sample_file.nc"
lon = -180:180
lat = -90:90
time = DateTime(2000,1,1):Day(1):DateTime(2000,1,3)
SST = randn(length(lon),length(lat),length(time))

ds = NCDataset(fname,"c")
defVar(ds,"lon",lon,("lon",));
defVar(ds,"lat",lat,("lat",));
defVar(ds,"time",time,("time",));
defVar(ds,"SST",SST,("lon","lat","time"));


# load by bounding box
v = @select(ds["SST"],30 <= lon <= 60 && 40 <= lat <= 80)

# substitute a local variable in condition using \$
lonr = (30,60) # longitude range
latr = (40,80) # latitude range

v = @select(ds["SST"],\$lonr[1] <= lon <= \$lonr[2] && \$latr[1] <= lat <= \$latr[2])

# You can also select based on `ClosedInterval`s from `IntervalSets.jl`.
# Both 30..60 and 60 ± 20 construct `ClosedInterval`s, see their documentation for details.

lon_interval = 30..60
lat_interval = 60 ± 20
v = @select(ds["SST"], lon ∈ \$lon_interval && lat ∈ \$lat_interval)

# get the indices matching the select query
(lon_indices,lat_indices,time_indices) = parentindices(v)

# get longitude matchting the select query
v_lon = v["lon"]

# find the nearest time instance
v = @select(ds["SST"],time ≈ DateTime(2000,1,4))

# find the nearest time instance but not earlier or later than 2 hours
# an empty array is returned if no time instance is present

v = @select(ds["SST"],time ≈ DateTime(2000,1,3,1) ± Hour(2))

close(ds)
```

Any 1D variable with the same dimension name can be used in `@select`. For example,
if we have a time series of temperature and salinity, the temperature values
can also be selected based on salinity:

```julia
# create a sample time series
using NCDatasets, Dates
fname = "sample_series.nc"
time = DateTime(2000,1,1):Day(1):DateTime(2009,12,31)
salinity = randn(length(time)) .+ 35
temperature = randn(length(time))

NCDataset(fname,"c") do ds
    defVar(ds,"time",time,("time",));
    defVar(ds,"salinity",salinity,("time",));
    defVar(ds,"temperature",temperature,("time",));
end

ds = NCDataset(fname)

# load all temperature data from January where the salinity is larger than 35.
v = @select(ds["temperature"],Dates.month(time) == 1 && salinity >= 35)

# this is equivalent to
v2 = ds["temperature"][findall(Dates.month.(time) .== 1 .&& salinity .>= 35)]

@test v == v2
close(ds)
```

!!! note

    For optimal performance, one should try to load contiguous data ranges, in
    particular when the data is loaded over HTTP/OPeNDAP.

"""
macro select(v,expression)
    expression_list = split_by_and(expression)
    args = Vector{Any}(undef,length(expression_list))

    # loop over all sub-expressions separated by &&
    for (i,e) in enumerate(expression_list)
        (param,newsym),e = scan_coordinate_name(e)

        if (e.head == :call) && (e.args[1] == :≈)
            @assert e.args[2] == newsym
            target = e.args[3]
            tolerance = nothing

            if (hasproperty(target,:head) &&
                (target.head == :call) && (target.args[1] == :±))
                value,tolerance = target.args[2:end]
            else
                value = target
            end

            args[i] = quote
                $(Meta.quot(param)) => Near($(esc(value)),$tolerance)
            end
        else
            fun = Expr(:->,newsym,e)
            args[i] = quote
                $(Meta.quot(param)) => $(esc(fun))
            end
        end
    end

    return Expr(:call,:select,esc(v),args...)
end

"""
    vsubset = CommonDataModel.select(v,param1 => condition1, param2 => condition2,...)
    dssubset = CommonDataModel.select(ds,param1 => condition1, param2 => condition2,...)

Return a subset of the variable `v` (or dataset `ds`) satisfying the conditions
applied to the corresponding parameters. `param1`, `param2` ... are symbols or
strings with variable names and `condition1`, `condition2` ... are functions
taking as a argument the values of the corresponding parameter and return
either false (ignore the corresponding elements) or true (select the
corresponding elements for loading).

## Examples

Create a sample file with random data:

```julia
using NCDatasets, Dates
using CommonDataModel: select, Near

fname = "sample_file.nc"
lon = -180:180
lat = -90:90
time = DateTime(2000,1,1):Day(1):DateTime(2000,1,3)
SST = randn(length(lon),length(lat),length(time));

ds = NCDataset(fname,"c")
defVar(ds,"lon",lon,("lon",));
defVar(ds,"lat",lat,("lat",));
defVar(ds,"time",time,("time",));
defVar(ds,"SST",SST,("lon","lat","time"));


# load by bounding box
v = select(ds["SST"],
           :lon => lon -> 30 <= lon <= 60,
           :lat => lat -> 40 <= lat <= 80)

# You can also select based on `ClosedInterval`s from `IntervalSets.jl`.
# Both 30..60 and 60 ± 20 construct `ClosedInterval`s, see their documentation for details. `∈` can be typed `\\in` followed by the TAB-key.

using IntervalSets
v = select(ds["SST"], :lon => ∈(30..60), :lat => ∈(60 ± 20))

# get the indices matching the select query
(lon_indices,lat_indices,time_indices) = parentindices(v)

# get longitude matchting the select query
v_lon = v["lon"]

# find the nearest time instance
v = select(ds["SST"],:time => Near(DateTime(2000,1,4)))

# find the nearest time instance but not earlier or later than 2 hours
# an empty array is returned if no time instance is present

v = select(ds["SST"],:time => Near(DateTime(2000,1,3,1),Hour(2)))

close(ds)
```

See also [`@select`](@ref CommonDataModel.@select) for more information.
"""
function select(v,conditions...)
    coord_names = coordinate_names(v)
    if v isa AbstractArray
        indices = Any[Colon() for _ in 1:ndims(v)]
    else
        indices = Dict{Symbol,Any}(((Symbol(d),Colon()) for d in dimnames(v)))
    end

    for (param,condition) in conditions
        coord,j = coordinate_value(v,param)

        if isa(condition,Near)
            ind = _findfirst(condition,coord)
            indices[j] = _intersect(indices[j],ind)
        else
            ind = findall(condition,coord)
            indices[j] = _intersect(indices[j],ind)
        end
    end

    if v isa AbstractArray
        view(v, indices...)
    else
        view(v; indices...)
    end
end

function coordinate_value(v::AbstractVariable,name_coord::Symbol)
    ncv = dataset(v)[name_coord]
    @assert ndims(ncv) == 1
    dimension_name = dimnames(ncv)[1]
    i = findfirst(==(dimension_name),dimnames(v))
    fmtd(v) = join(dimnames(v),"×")

    if i == nothing
        error("$name_coord (dimensions: $(fmtd(ncv))) and $(name(v)) (dimensions: $(fmtd(v))) do not share a named dimension")
    end
    return Array(ncv),i
end


function coordinate_names(v::AbstractVariable)
    ds = dataset(v)
    dimension_names = dimnames(v)

    return [Symbol(varname) for (varname,ncvar) in ds
     if (ndims(ncvar) == 1) && dimnames(ncvar) ⊆ dimension_names]
end


function coordinate_value(ds::AbstractDataset,name_coord::Symbol)
    ncv = ds[name_coord]
    @assert ndims(ncv) == 1
    return Array(ncv),Symbol(dimnames(ncv)[1])
end

function coordinate_names(ds::AbstractDataset)
    return [Symbol(varname) for (varname,ncvar) in ds
       if (ndims(ncvar) == 1)]
end


#  LocalWords:  params vsubset conditionN NetCDF coord NCDatasets lon
#  LocalWords:  julia dereferenced parentindices fname nc DateTime ds
#  LocalWords:  randn NCDataset defVar lonr latr CFTime OPeNDAP args
#  LocalWords:  hasproperty esc Expr fmtd ncv
